(*----------------------------------------------------------*)
(*                                                          *)
(*  WINDOW.DEF - Modula-2 window Partial lib                *)
(*                                                          *)
(*  Colors and clipping not implemented.                    *)
(*  Modified for Topspeed Clarion V5.0   CWS    12-08-00    *)
(*                                                          *)
(*----------------------------------------------------------*)

(*#call(o_a_copy=>off) *)
(*    # check(stack=>off,
          index=>off,
          range=>off,
          overflow=>off,
          nil_ptr=>off) *)

IMPLEMENTATION MODULE Window;

IMPORT Str, Lib, IO, CWRUN, M2SCROLL;               (* 04-14-99 *)
FROM CWRUN IMPORT SIGNED;

CONST MaxWinArray = 6;
VAR
  WinArray: ARRAY[0..MaxWinArray] OF WinDescriptor;    (* fudge allocation *)  (* 05-07-99 *)
  WinArrayPtr: CARDINAL;
  ScreenList: WinType;                             (* 05-10-99 *)
  CurrentScreen, LastRefresh: WinType;             (* 05-13-99 *)



 PROCEDURE  CurWin(): WinType;
   (* Returns The current window being used for output for this process *)
   (* If no window assigned by Use then returns Top.                    *)
   (* NB Locks window system and leaves locked if _mthread set.         *)
   BEGIN
     IF CurrentScreen = NIL THEN RETURN FullScreen;
     ELSE RETURN CurrentScreen;
     END;
   END CurWin;




PROCEDURE  DrawFrame ( W : WinType );                     (* 05-06-99 *)
VAR
  s      : ARRAY[0..81] OF CHAR;
  tl     : CARDINAL;
  i      : CARDINAL;

  PROCEDURE ScreenWrite(W: WinType; Xc,Yc: RelCoord; St: ARRAY OF CHAR);   (* 05-11-99 *)
    VAR Slen: CARDINAL;
    BEGIN    (* Writes to outer frame without scroll *)
      Slen := Str.Length(St);
      IF Slen > 0 THEN
        WITH W^.SC DO
        WITH W^.WDef DO
          CWRUN.BLANK( VAL(SIGNED,XF*(X1+Xc)+M2SCROLL.Hbdr), VAL(SIGNED,YF*(Y1+Yc)+M2SCROLL.Vbdr),  (* 05-12-99 *)
                         VAL(SIGNED,XF*Slen), VAL(SIGNED,YF+1) );                                       (* 05-12-99 *)
          CWRUN.Show( VAL(SIGNED,XF*(X1+Xc)+M2SCROLL.Hbdr), VAL(SIGNED,YF*(Y1+Yc)+M2SCROLL.Vbdr), St );  (* 05-12-99 *)
        END; (* with *)
        END; (* with *)
      END;
    END ScreenWrite;


  PROCEDURE PutTitle( mode : CARDINAL; row : CARDINAL ); (* Private *)
  VAR
    i,j : CARDINAL;
  BEGIN
    WITH W^ DO
      CASE mode OF
       | 0 : i := 1;
       | 1 : i := (OWidth-tl) DIV 2;                 (* 05-12-99 *)
       | 2 : i := OWidth-tl;                         (* 05-12-99 *)
      ELSE
        i := MAX(CARDINAL);
      END;
      j := 0;
      WHILE (i<=OWidth-1) AND (j<tl) DO
        s[i] := Title[j]; INC(i); INC(j);            (* 05-06-99 *)
      END;
      ScreenWrite(W, 0, row, s);                     (* 05-07-99 *)
    END;
  END PutTitle;

BEGIN    (* DrawFrame *)
  WITH W^ DO
    IF NOT WDef.FrameOn THEN RETURN END;                (* 05-06-99 *)
    IF (OWidth<3)OR(ODepth<3) THEN RETURN END; (* No Room *)  (* 05-06-99 *)
    (* Top *)
    s[0] := WDef.FrameDef[0];
    Lib.Fill(ADR(s[1]),OWidth-2,WDef.FrameDef[1]);      (* 05-11-99 *)
    s[OWidth-1] := WDef.FrameDef[2];                    (* 05-11-99 *)
    IF (OWidth-1) < HIGH(s) THEN                        (* 05-11-99 *)
      s[OWidth(*+1-1*)] := CHR(0);                      (* 05-11-99 *)
    END;
    IF TMode # NoTitle THEN
      tl := Str.Length(Title);                          (* 05-06-99 *)
      IF tl>(OWidth-2) THEN tl := OWidth-2; END;        (* 05-12-99 *)
    END;
    PutTitle(ORD(TMode-LeftUpperTitle),0);
    (* Sides *)
    FOR i := 1 TO ODepth-1 DO                           (* 05-12-99 *)
      ScreenWrite(W, 0, i, WDef.FrameDef[3]);           (* 05-07-99 *)
      ScreenWrite(W, OWidth-1, i, WDef.FrameDef[4]);    (* 05-11-99 *)
    END;
    (* Bottom *)
    s[0] := WDef.FrameDef[5];
    Lib.Fill(ADR(s[1]),OWidth-2,WDef.FrameDef[6]);
    s[OWidth-1] := WDef.FrameDef[7];                     (* 05-11-99 *)
    IF (OWidth-1) < HIGH(s) THEN                         (* 05-11-99 *)
      s[OWidth(*+1-1*)] := CHR(0);                       (* 05-11-99 *)
    END;
    PutTitle(ORD(TMode-LeftLowerTitle),ODepth-1);        (* 05-12-99 *)
  END;
END DrawFrame;



PROCEDURE Refresh( W: WinType );  (* private *)          (* 05-13-99 *)
  BEGIN
    IF NOT W^.WDef.Hidden THEN
      LastRefresh := W;
      DrawFrame(W);
      M2SCROLL.Refresh(W^.SC);
    END;
  END Refresh;




(* ------------------------- *)
(* Cursor Control            *)
(* ------------------------- *)



PROCEDURE  CursorOn();
BEGIN
END CursorOn;

PROCEDURE  CursorOff();
BEGIN
END CursorOff;


(* ------------------------- *)
(* Window creation           *)
(* ------------------------- *)


PROCEDURE  Open ( WD : WinDef ) : WinType;
(*
  Opens a window on the screen ready for use
*)
CONST min = 2;              (* 07-14-00 *)
VAR W: WinType;
BEGIN
  IF WinArrayPtr > MaxWinArray THEN Lib.FatalError("No WinArray slot in Window.mod"); END;
  W := ADR(WinArray[WinArrayPtr]);                    (* 05-07-99 *)
  INC(WinArrayPtr);                                   (* 05-07-99 *)
  WITH W^ DO
    Next := NIL;                                      (* 05-10-99 *)
    WDef := WD;
    WITH WDef DO                                      (* 05-11-99 *)
      IF X2 >= ScreenWidth THEN X2:=ScreenWidth-1 END;
      IF Y2 >= ScreenDepth THEN Y2:=ScreenDepth-1 END;
     (* IF Y2 >= CurrentScreenDepth THEN Y2:=CurrentScreenDepth-1 END; *)
      IF FrameOn THEN                                 (* 05-11-99 *)
        IF (X1+min>X2) THEN X2 := X1+min END;             (* 05-06-99 *)
        IF (Y1+min>Y2) THEN Y2 := Y1+min END;             (* 05-06-99 *)
        M2SCROLL.Open(SC, X1+1, Y1+1, X2-1, Y2-1);        (* 05-11-99 *)
      ELSE
        (* min := 0; *)
        IF (X1>X2) THEN X2 := X1; END;                    (* 05-10-99 *)
        IF (Y1>Y2) THEN Y2 := Y1; END;                    (* 05-10-99 *)
        M2SCROLL.Open(SC, X1, Y1, X2, Y2);                (* 05-07-99 *)
      END;
      OWidth := X2-X1+1; ODepth := Y2-Y1+1;               (* 05-12-99 *)
    END; (* with *)
    TMode := NoTitle;                                   (* 05-06-99 *)
    Title := "";                                        (* 05-06-99 *)
  END; (* with *)
  IF WD.Hidden THEN                                     (* 05-12-99 *)
    Use ( W )
  ELSE
    PutOnTop ( W )
  END; 
  RETURN W;                                           (* 05-06-99 *)
END Open;


(* ------------------------- *)
(* Window stack manipulation *)
(* and screen redraw         *)
(* ------------------------- *)



PROCEDURE  Use ( W : WinType );
(*
   Causes all subsequent output (by the current process)
   to appear in the specified Window
   NB does not have to be Top Window (or in fact on the screen at all)
*)
BEGIN
  CurrentScreen := W;                                   (* 05-06-99 *)
END Use;



PROCEDURE  PutOnTop ( W : WinType );
(*
   Puts the specified window on the top of the window stack
   Ensuring that it is fully visible.
*)
VAR ScrPtr: WinType;                     (* 05-10-99 *)
BEGIN
  IF ScreenList = NIL THEN                 (* 05-10-99 *)
    ScreenList := W;                       (* 05-10-99 *)
    W^.Next := NIL;                        (* 05-10-99 *)      
  ELSE
    ScrPtr := ScreenList;                  (* 05-10-99 *)
    (* First block is fullscreen *)
    WHILE ScrPtr^.Next # NIL DO            (* 05-10-99 *)
      IF ScrPtr^.Next = W THEN             (* 05-10-99 *)
        (* Unlink us *)
        ScrPtr^.Next := W^.Next;           (* 05-10-99 *)
        W^.Next := NIL;                    (* 05-10-99 *)      
      ELSE
        ScrPtr := ScrPtr^.Next;            (* 05-10-99 *)
      END;
    END;
    (* Link at top *)
    ScrPtr^.Next := W;                     (* 05-10-99 *)
  END;
  W^.WDef.Hidden := FALSE;                 (* 05-12-99 *)
  Use ( W );                               (* 05-12-99 *)
  Refresh( W );                            (* 05-13-99 *)
END PutOnTop;


PROCEDURE  Hide ( W : WinType );
(*
    Removes window from the Window stack and also the screen
    Uncovers obscured windows
*)
VAR ScrPtr: WinType;                     (* 05-11-99 *)
BEGIN
  W^.WDef.Hidden := TRUE;                (* 05-06-99 *)
  IF ScreenList = NIL THEN RETURN; END;  (* 05-10-99 *)
  ScrPtr := ScreenList;                  (* 05-10-99 *)
  (* First block is fullscreen *)
  WHILE ScrPtr # NIL DO                  (* 05-10-99 *)
    IF ScrPtr^.Next = W THEN             (* 05-10-99 *)
      (* Unlink us *)
      ScrPtr^.Next := W^.Next;           (* 05-10-99 *)
      W^.Next := NIL;                    (* 05-10-99 *)      
    END;
    Refresh(ScrPtr);                     (* 05-13-99 *)
    IF ScrPtr # NIL THEN                   (* 05-10-99 *)
      ScrPtr := ScrPtr^.Next;              (* 05-10-99 *)
    END;
  END;
END Hide;


PROCEDURE  PutBeneath ( W : WinType; WA : WinType );
(*
    Puts window W beneath window WA
*)
VAR ScrPtr: WinType;  AboveW: BOOLEAN;   (* 05-10-99 *)
BEGIN
  Hide(W);  (* Unlink us *)
  AboveW := FALSE;
  IF ScreenList = NIL THEN RETURN; END;  (* 05-10-99 *)
  ScrPtr := ScreenList;                  (* 05-10-99 *)
  (* First block is fullscreen *)
  WHILE ScrPtr^.Next # NIL DO            (* 05-10-99 *)
    IF AboveW THEN                       (* 05-10-99 *)
      Refresh(ScrPtr);                   (* 05-13-99 *)
    END;
    IF ScrPtr^.Next = WA THEN            (* 05-10-99 *)
      (* Relink us *)
      ScrPtr^.Next := W;                 (* 05-10-99 *)
      W^.Next := WA;                     (* 05-10-99 *)      
      AboveW := TRUE;
    END;
    ScrPtr := ScrPtr^.Next;              (* 05-10-99 *)
  END;
END PutBeneath;


PROCEDURE  SnapShot();
(* Updates the Window buffer from the previous window. *)
VAR Ix, Iy: CARDINAL;                               (* 05-07-99 *)
    Cptr, Lptr: POINTER TO M2SCROLL.ScWinDesc;      (* 05-10-99 *)
    ScrPtr: WinType;                                (* 05-10-99 *)
BEGIN
  IF ScreenList = NIL THEN RETURN; END;  (* 05-10-99 *)
  ScrPtr := ScreenList;                  (* 05-10-99 *)
  Lptr := NIL;
  Cptr := ADR(CurrentScreen^.SC);        (* 05-10-99 *)
  (* First block is fullscreen *)
  LOOP
    IF ScrPtr^.Next = NIL THEN EXIT; END;   (* 05-13-99 *)
    IF ScrPtr^.Next = CurrentScreen THEN    (* 05-10-99 *)
      EXIT;  (* Take previous *)            (* 05-13-99 *)
    END;
    ScrPtr := ScrPtr^.Next;              (* 05-10-99 *)
  END;
  (* If hide current screen then take top *) 
  Lptr := ADR(ScrPtr^.SC);               (* 05-13-99 *)
 (* NOTE: Not allowing for different size windows. *)
  IF (Cptr^.Depth # Lptr^.Depth) OR (Cptr^.Width # Lptr^.Width) THEN
    RETURN;
  END;
  Cptr^.TTYptr := Lptr^.TTYptr;          (* 05-07-99 *)
  FOR Ix := 0 TO Cptr^.Depth-1 DO        (* 05-07-99 *)
    Str.Slice(Cptr^.TTYstring[Ix], Lptr^.TTYstring[Ix], 0, Cptr^.Width);  (* 05-07-99 *)
  END; (* for *)
  Refresh( CurrentScreen );                 (* 05-13-99 *)
END SnapShot;




PROCEDURE  Close ( VAR W : WinType );
(*
     removes the specified window from the screen
     deletes window descriptor and
     de-allocates any buffers previously allocated.
*)
BEGIN
  Hide ( W );  (* Unlink us *)               (* 05-06-99 *)
  W^.OWidth := 0;                            (* 12-08-00 *)
  W := NIL;                                  (* 05-10-99 *)
  WHILE (WinArrayPtr > 0) AND (WinArray[WinArrayPtr-1].OWidth = 0) DO  (* 11-21-00 *)
    DEC(WinArrayPtr);                        (* 05-10-99 *)
  END; (* while *)
END Close;




PROCEDURE  Used () : WinType;
(*
    Returns The current window being used for output for this process
    If no window assigned by Use then returns Top
*)
VAR
  w : WinType;
BEGIN
  w := CurWin();
  RETURN w;
END Used;


PROCEDURE  Top () : WinType;
(*
    Returns The current top Window
*)
VAR ScrPtr: WinType;                     (* 05-10-99 *)
BEGIN
  IF ScreenList = NIL THEN RETURN NIL; END;  (* 05-10-99 *)
  ScrPtr := ScreenList;                  (* 05-10-99 *)
  (* First block is fullscreen *)
  WHILE ScrPtr^.Next # NIL DO            (* 05-10-99 *)
    ScrPtr := ScrPtr^.Next;              (* 05-10-99 *)
  END;
  RETURN ScrPtr;
END Top;



(* ------------------------- *)
(* Window output             *)
(* ------------------------- *)



PROCEDURE  WriteOut (S : ARRAY OF CHAR);
VAR
  W : WinType;
BEGIN
  W := CurWin();
  IF LastRefresh # W THEN
    Refresh( W );
  END;
  WITH W^ DO
    M2SCROLL.WrStr(SC, S, WDef.WrapOn, WDef.Hidden);            (* 05-12-99 *)
  END; (* with *)
END WriteOut;



PROCEDURE  Clear();
(*
  clears the current window
*)
VAR
  w : WinType;
BEGIN
  w := CurWin();
  M2SCROLL.Clear( w^.SC );             (* 05-07-99 *)
  Refresh( w );                        (* 05-13-99 *)
END Clear;


PROCEDURE  ClrEol();
(*
   clears from the cursor to the end of line
*)
BEGIN
END ClrEol;


PROCEDURE  GotoXY ( X,Y : RelCoord );
BEGIN
  M2SCROLL.GoToXY(CurWin()^.SC, X, Y);      (* 05-07-99 *)
END GotoXY;


PROCEDURE  WhereX ( ) : RelCoord;
VAR
  W : WinType;
BEGIN
  W := CurWin();
  RETURN W^.SC.CurrentX;
END WhereX;


PROCEDURE  WhereY ( ) : RelCoord;
VAR
  W : WinType;
BEGIN
  W := CurWin();
  RETURN W^.SC.CurrentY;
END WhereY;


PROCEDURE  TextColor ( c : Color );
BEGIN
END TextColor;

PROCEDURE  TextBackground ( c : Color );
BEGIN
END TextBackground;

PROCEDURE  SetWrap ( on : BOOLEAN );
BEGIN
END SetWrap;




(* ------------------------- *)
(* Title procedure           *)
(* ------------------------- *)
PROCEDURE SetTitle ( W        : WinType;
                     NewTitle : ARRAY OF CHAR;
                     Mode     : TitleMode );
(*
  updates the window title within the window frame,
  positioning it in the position defined by the title mode
*)
BEGIN
  WITH W^ DO
    IF Mode # NoTitle THEN                           (* 05-06-99 *)
      Str.Copy(Title, NewTitle);                     (* 05-06-99 *)
    END;
    TMode := Mode;                                   (* 05-06-99 *)
  END;
  IF NOT W^.WDef.Hidden THEN                         (* 05-10-99 *)
    DrawFrame(W);
  END;
END SetTitle;



PROCEDURE ReadString  ( VAR string : ARRAY OF CHAR );
VAR
  c   : CHAR;
  line: ARRAY[0..82] OF CHAR;
  p,H : CARDINAL;
  W   : WinType;
  con : BOOLEAN;
BEGIN
  W := CurWin();
  PutOnTop(W);
  con := TRUE;    (* 04-14-99 *)
  CursorOn();
  H := HIGH(string);
  IF H>79 THEN H := 79 END;
  p := 0;
  LOOP
    c := IO.RdKey();
    IF (c=CHR(8))OR(c=CHR(127)) THEN
      IF p>0 THEN DEC(p); IO.WrChar(CHR(8)) END;
    ELSIF (c>=' ') THEN
      IF p<=H THEN
        IO.WrChar(c);
        line[p] := c;
        INC(p);
      END;
    ELSIF c=CHR(13) THEN
      EXIT;
    END;
  END;
  line[p] := CHR(0);
  Str.Copy(string,line);
  IF NOT con THEN CursorOff END;
  IO.WrLn();
END ReadString;



(* ------------------------- *)
(* Main initialization       *)
(* ------------------------- *)

CONST
  ClearOnEntry = TRUE;    (* Change to FALSE if automatic clear *)


BEGIN
  FOR WinArrayPtr := 0 TO MaxWinArray DO      (* 11-21-00 *)
    WinArray[WinArrayPtr].OWidth := 0;        (* 11-21-00 *)
  END; (* for *)
  WinArrayPtr := 0;         (* 04-14-99 *)
  ScreenList := NIL;                      (* 05-10-99 *)
  LastRefresh := NIL;                     (* 05-13-99 *)
  IO.WrStrRedirect := WriteOut;
  IO.RdStrRedirect := ReadString;
  IF ClearOnEntry THEN                    (* 05-11-99 *)
    FullScreen := Open(FullScreenDef);
  ELSE
    FullScreen^.WDef.Hidden := TRUE;      (* 05-07-99 *)
    FullScreen := Open(FullScreenDef);
    SnapShot();  (* get previous text, N/A in MSwin *)
    PutOnTop(FullScreen);
  END;

END Window.
